# ==================================================================
# Gossamer Threads Module Library - http://gossamer-threads.com/
#
#   GT::HTML
#   Author  : Aki Mimoto
#   $Id: HTML.pm,v 1.13 2002/04/19 17:29:38 aki Exp $
#
# Copyright (c) 2000 Gossamer Threads Inc.  All Rights Reserved.
# ==================================================================
#
# Description:
#   Module that does HTML parsing
#

package GT::HTML;

use strict;
use GT::Base;
use GT::MD5 qw( md5_hex );
use Exporter;
use vars     qw/ $ATTRIBS @ISA $ENTITY_VALUE @EXPORT /;

@ISA = qw| Exporter GT::Base |;

@EXPORT = qw| &_clean_path |;

$ENTITY_VALUE = {
                'quot' => 34,
                'amp' => 38,
                'lt' => 60,
                'gt' => 62,
                'nbsp' => 160,
                'iexcl' => 161,
                'cent' => 162,
                'pound' => 163,
                'curren' => 164,
                'yen' => 165,
                'brvbar' => 166,
                'sect' => 167,
                'uml' => 168,
                'copy' => 169,
                'ordf' => 170,
                'laquo' => 171,
                'not' => 172,
                'shy' => 173,
                'reg' => 174,
                'macr' => 175,
                'deg' => 176,
                'plusmn' => 177,
                'sup2' => 178,
                'sup3' => 179,
                'acute' => 180,
                'micro' => 181,
                'para' => 182,
                'middot' => 183,
                'cedil' => 184,
                'sup1' => 185,
                'ordm' => 186,
                'raquo' => 187,
                'frac14' => 188,
                'frac12' => 189,
                'frac34' => 190,
                'iquest' => 191,
                'agrave' => 192,
                'aacute' => 193,
                'acirc' => 194,
                'atilde' => 195,
                'auml' => 196,
                'aring' => 197,
                'aelig' => 198,
                'ccedil' => 199,
                'egrave' => 200,
                'eacute' => 201,
                'ecirc' => 202,
                'euml' => 203,
                'igrave' => 204,
                'iacute' => 205,
                'icirc' => 206,
                'iuml' => 207,
                'eth' => 208,
                'ntilde' => 209,
                'ograve' => 210,
                'oacute' => 211,
                'ocirc' => 212,
                'otilde' => 213,
                'ouml' => 214,
                'times' => 215,
                'oslash' => 216,
                'ugrave' => 217,
                'uacute' => 218,
                'ucirc' => 219,
                'uuml' => 220,
                'yacute' => 221,
                'thorn' => 222,
                'szlig' => 223,
                'agrave' => 224,
                'aacute' => 225,
                'acirc' => 226,
                'atilde' => 227,
                'auml' => 228,
                'aring' => 229,
                'aelig' => 230,
                'ccedil' => 231,
                'egrave' => 232,
                'eacute' => 233,
                'ecirc' => 234,
                'euml' => 235,
                'igrave' => 236,
                'iacute' => 237,
                'icirc' => 238,
                'iuml' => 239,
                'eth' => 240,
                'ntilde' => 241,
                'ograve' => 242,
                'oacute' => 243,
                'ocirc' => 244,
                'otilde' => 245,
                'ouml' => 246,
                'divide' => 247,
                'oslash' => 248,
                'ugrave' => 249,
                'uacute' => 250,
                'ucirc' => 251,
                'uuml' => 252,
                'yacute' => 253,
                'thorn' => 254,
                'yuml' => 255
            };

$ATTRIBS    = {
                documents => {
#                                           'uri'   => {
#                                                       'request'   => GT::URI::HTTP,
#                                                       'buffer'    => ''
#                                                       }
                            }
            };


################################################################################
################################################################################
################################################################################

sub parse_documents {
#-------------------------------------------------------------------------------
# takes document(s) and parses them
    my $self        = shift;

# first figure out how we're trying to rack this
    my $document;
    if ( ref $_[0] eq 'HASH' ) {
        $document   = shift;
    }

    elsif ( not ( scalar(@_) % 2 ) ) {
        my %tmp     = @_;
        $document   = { map {
                            $_ => {
                                    'buffer'            => $tmp{$_},
                                    'resource_attribs'  => {
                                                            'content-type'  => 'text/html'
                                                            }

                                    }
                            } keys %tmp };
    }

    elsif ( scalar(@_) == 1 ) {
        $document   = {
                        'user'  =>
                                    {
                                    'buffer'            => shift,
                                    'resource_attribs'  => {
                                                            'content-type'  => 'text/html'
                                                            }
                                    }
                        }
    }

    else {
        return undef;
    }

# now rack the requested documents
    foreach my $uri ( keys %{$document} ) {
        $self->{documents}->{$uri} = $document->{$uri};
    }
    return $self->_parse_documents();
}

sub _parse_documents {
#-------------------------------------------------------------------------------
# takes all the unparsed uris and parses them
    my $self        = shift;
    my $documents   = $self->{documents} or return $self->debug("GT::HTML->{documents} is null? something's wrong, what's up?");
    my @parse_uris  = grep { 
                                defined( $documents->{$_}->{buffer} ) 
                                and 
                                not defined( $documents->{$_}->{parsed_data} )
                            } 
                            keys %{$documents};

# for each url that has been found to be not parsed, parse them
    foreach my $uri ( @parse_uris ) {
        my $URI     = $documents->{$uri};
        my $attribs = $URI->{resource_attribs} or next;

# ... only parse if we're dealing with a text/html page
        if ( lc($attribs->{"content-type"}) !~ m,text/html,i ) { next; }

# ... now parse
        $URI->{parsed_data} = $self->_parse_document( $URI->{buffer} );

# ... dereference those relative URLs!
        _dereference_hrefs( $uri, $URI->{parsed_data} );
    }

}

sub clear_documents {
#-------------------------------------------------------------------------------
# clears the document cache
    my $self        = shift;

    $ATTRIBS->{documents} = {};
    $self->{documents} = {};
}

################################################################################
################################################################################
################################################################################

sub _parse_document {
#-------------------------------------------------------------------------------
# takes a document reference and parses it
    my $self        = shift;
    my $document    = shift  or return $self->debug("Invalid Arguments");

# split the document into <tag>,document,<tag>,<tag> that can be parsed
    my @tokenised   = split /(
                                < 
                                    (?: 
                                        # no wierd chars...
                                        [^">] 

                                        # with " quote escapes...,
                                        | "+(?:[^">]|\\")*?"+? 

                                        # with ' quote escapes
                                        | '+(?:[^'>]|\\')*?'+?

                                        )*? 
                                >
                            )/xso, $document;

# no go through and break down the document into pieces
#
# State 0:
#           normal, text is normal, no special modifier tags
#
# State 1:
#           link, text is link information
#
# State 2: 
#           javascript, or something else, ignore until stop request has been received
#
# State 3:
#           title of document
#
    my %parse   = ();
    my @state_stack = (
#                       {
#                           state => 0,
#                           attribs => { specials }
#
#                       }
                    );
    my $state = {
                    state => 0,
                    attribs => {}
                };

    foreach my $token ( @tokenised ) {
        chomp $token;

        $self->debug( "STATE: $state->{state}" ) if ($self->{_debug});

# ... token is a tag
        if ( substr( $token, 0, 1 ) eq '<' ) {
            $self->debug( "Found Tag: `$token'", 3 ) if ($self->{_debug});

# .... strip the '<', '>' from the quotes
            my $parameters  = substr $token, 1, -1 or next;

# .... parse the tag
            my $tag_info    = _tag_parse( $parameters );
            my $tag_attribs = $tag_info->{tags};
            my $tag_type    = $tag_info->{type};

            # we're not in a mode where we're skipping due to javascript and stuff are we?
            if ( 
                    $state->{state} == 2 
                    and ( $tag_type ne 'script' )
                    and ( not $tag_info->{closed} )
                    ) {
                    $self->debug( "Skipping $token: ( $tag_type | $tag_info->{closed})" ) if ($self->{_debug});
                    next;
            }

            elsif (
                    $state->{state} == 4 
                    and ( $tag_type ne 'style' )
                    and ( not $tag_info->{closed} )
                    ) {
                    $self->debug( "Skipping $token: ( $tag_type | $tag_info->{closed})" ) if ($self->{_debug});
                    next;
            }

        # Is a link
            elsif ( $tag_type eq 'a' ) {

        # ... if we are trying to close
                if ( $tag_info->{close} and $state->{state} == 1) {
                    push @{$parse{a}}, {
                                            attribs => $state->{attribs},
                                            text    => $state->{text}
                                            };
                    $state = pop @state_stack || {
                                                    state   => 0
                                                    };
                }

        # ... are we currently open with a link
                elsif ( $state->{state} == 1 ) {

                    push @{$parse{a}}, {
                                            attribs => $state->{attribs},
                                            text    => $state->{text}
                                            };
                    $state = pop @state_stack;
                    $state = {
                                state   => 1,
                                attribs => $tag_attribs,
                                text    => ''
                                };

                    
                }

                elsif ( $state->{state} == 0 ) {
                    $state = {
                                state   => 1,
                                attribs => $tag_attribs,
                                text    => ''
                                };
                }
            }

        # Image
            elsif ( $tag_type eq 'img' ) {
                push @{$parse{img}}, $tag_attribs;
            }

        # Javascript
            elsif ( $tag_type eq 'script' ) {
                if ( $tag_info->{close}  ) {
                    if ( $state->{state} == 2 ) {
                        $state = pop @state_stack;
                    }
                }

                else {
                    push @state_stack, { %{$state} };
                    $state = {
                                state   => 2,
                                attribs => {}
                                };
                }
            }

        # Title
            elsif ( $tag_type eq 'title' ) {

                if ( $tag_info->{close} ) {
                    if ( $state->{state} == 3) {
                        $parse{title}   = $state->{title};
                        $state = pop @state_stack;
                    }
                }

                else {
                    push @state_stack, { %{$state} };
                    $state = {
                                state   => 3,
                                attribs => {}
                            };
                }

            }

        # Meta
            elsif ( $tag_type eq 'meta' and !$tag_info->{close}) {
                push @{$parse{meta}}, $tag_attribs;
            }

        # Stylesheet
            elsif ( $tag_type eq 'style' ) {
                if ( $tag_info->{close}  ) {
                    if ( $state->{state} == 4 ) {
                        $state = pop @state_stack;
                    }
                }

                else {
                    push @state_stack, { %{$state} };
                    $state = {
                                state   => 4,
                                attribs => {}
                                };
                }
            }

        }

# ... or a piece of text?
        else {
            $self->debug( "Found Text: `$token'", 3 ) if ($self->{_debug});
            my $tmp             = &de_htmlize($token) . " ";

            # normal
            if ( $state->{state} == 0 ) {
                $parse{'bare_text'} .= $tmp;
            }

            # link
            elsif ( $state->{state} == 1 ) {
                $parse{'bare_text'} .= $tmp;
                $state->{text}      .= $tmp;
            }

            # javascript
            elsif ( $state->{state} == 2 ) {
                # do nothing
            }

            # title 
            elsif ( $state->{state} == 3 ) {
# uncomment the following line if title is considered to be part of the text as well
#               $parse{'bare_text'} .= $tmp;
                $state->{title}     .= $tmp;
            }
        }

    }

# finish up by dropping all the extra spaces foundin the bare text area
    $parse{'bare_text'} =~ s/(\s+)/ /g;

# calculate the MD5 value of the text
    $parse{'digest'}    = md5_hex($parse{'bare_text'} || $document);

    return \%parse;

}

sub _dereference_hrefs {
#-------------------------------------------------------------------------------
# takes all the current object's 'a hrefs' and parses them out
    my $base_uri    = shift;
    my $parsed_data = shift;
    my $arefs       = $parsed_data->{'a'};
    my @uris        = map { $_->{attribs}->{'href'} } @{$arefs};
    my $deref_uris  = _dereference_urls( $base_uri, @uris );

    foreach my $index ( 0 .. $#{$arefs} ) {
        $arefs->[$index]->{attribs}->{href} = $deref_uris->[$index];
    }

}

sub _dereference_urls {
#-------------------------------------------------------------------------------
# takes a url and makes it complete
# assumes HTTP/1.0
    my $base_uri    = shift;
    my @uris        = @_;
    my @deparsed    = ();

# deparse the base host
    $base_uri   =~ m,\w+://(.*)\??,i;
    my $dirs    = _fragment_path( $1 );
    my $host    = shift @{$dirs};
    my $page    = pop @{$dirs};

# now start parsing each of the urls
    foreach my $uri ( @uris ) {

        if ( $uri =~ m,^\s*\w+:,  ) {
            push @deparsed, $uri;
        }

        else {
            my $rec_path = _fragment_path( $uri );
            my $new_path = $rec_path->[0] ? _merge_paths( $dirs, $rec_path ) : $rec_path;
            push @deparsed, "http://$host/" . ( join('/', @{__clean_path($new_path) || [] } ) || '' );
        }
    }

    return \@deparsed;

}

sub _clean_path {
#-------------------------------------------------------------------------------
# takes a single URI and tries to clean it up
    my $uri     = shift;
    $uri =~ m,\w+://(.*)\??,i;
    my $dirs    = _fragment_path( $1 );
    my $host    = shift @{$dirs};
    return "http://$host/" . ( join('/', grep defined $_, @{ __clean_path( $dirs ) || [] } ) );
}

sub _fragment_path {
#-------------------------------------------------------------------------------
# fragments into '/' spaced stuff
    my $path    = shift;

    my @dirs    = ();
    my $i       = 0;
    my $j       = 0;

    while ( ( $i = index( $path, '/', $i ) ) >= 0 ) {
        my $tmp = substr( $path, $j, ($i-$j) );
        push @dirs, $tmp;
        $i++;
        $j = $i;
    }

    push @dirs, substr( $path, $j );

    return \@dirs;
}

sub __clean_path {
#-------------------------------------------------------------------------------
# makes sure that we don't get crap like ///// going on
    my $path    = shift or return undef;

    my @dirs    = ();
    my $page    = pop @{$path};
    @dirs       = grep $_, @{$path};
    push @dirs, $page;
    return \@dirs;
}

sub _merge_paths {
#-------------------------------------------------------------------------------
# merges two paths together
    my $base    = shift;
    my $new     = shift;

# now for each of the new url bases, simulate the file system, back up one, don't do anything, etc
    my @dirs    = @{$base};
    foreach my $dir ( @{$new} ) {

        if ( $dir eq '..' ) {
            pop @dirs;
        }

        elsif ( $dir eq '.' ) {
        }

        else {
            push @dirs, $dir;
        }
    
    }

    return \@dirs;

}

# Copyright (c) 2000 Mel
sub _tag_parse {
#-------------------------------------------------------------------------------
# takes the stuff in between tags and turns them into a hash of name -> value
# if it's just a unvalued item, say 'b' it will be undef'd but will exist
    my $string = shift;
    $string =~ s/\s+=/=/g;
    $string =~ s/=\s+/=/g;
    $string =~ s|^/\s+|/|;
    my @args = grep /\S/, split /([^=\s]+(?:=(?:"[^"]*"|'[^']*'|\S+))?)/, $string;
    my $type = shift @args;
    my $attrib = {};
    for (@args) {
    my ($k,$v) = split /=/,$_,2;
    if (defined $v) {
        unless ($v =~ s/^'// and $v =~ s/'$//) {
            $v =~ s/^"// and $v =~ s/"$//;
        }
    }
    $attrib->{lc $k} = $v;
    }
    my $close = $type =~ s|^/||;
    return { type => lc $type, tags => $attrib, close => $close };
}


sub de_htmlize {
#-------------------------------------------------------------------------------
# takes those pesky &quot;metacharacter thingies&quot;
    my $text    = shift;

    $text =~ s/&(\#?)([a-zA-Z0-9]+);/
                                    $1 ? chr( $2 ) : ( chr($ENTITY_VALUE->{lc $2}) || "&$1$2;" )
                                    /gex;
    
    return $text;
}

sub htmlize {
#-------------------------------------------------------------------------------
# takes a normal string and html encodes with all the appropriate &123; codes 
#
    my $text    = shift;

    $text =~ s/(["'<>()&])/
                        "&#" . ord( $1 ) . ";"
                        /gex;

    return $text;
}

sub parse {
#-------------------------------------------------------------------------------
# a quick parsing technique
    my $ref         = shift; # get rid of the initial argument

    my $documents   = shift;
    my $options     = shift;

    my $HTML        = ( ref $ref eq 'GT::HTML' ) ?  $ref : new GT::HTML( $options );
    $HTML->parse_documents($documents);
    $documents      = $HTML->documents();
    $HTML->clear_documents();

    return $documents;
}


1;

__END__

=head1 NAME

GT::HTML - A HTML file parser

=head1 SYNOPSIS

  use GT::HTML;
  use GT::URI;
  use GT::Dumper;

  # get some html data
  my $doc = GT::URI->get( "http://www.yahoo.com" );

  # parse it out..
  my $parsed = GT::HTML->parse( $doc );

  # display the parsed out document
  print Dumper($parsed);

=head1 BASICS

=head2 Breaking down an HTML document the easy way

The most important thing to keep in mind about this module, is that it's designed to work with GT::URI. As well, the module is designed to extract data, not analyse the structure.

The synopsis gives the simplest example, however, if there is an already existing document in a buffer you'd like to parse.

  use GT::HTML;
  use GT::Dumper;

  # get data
  open FILE, $ARG[0] or die "$!";
  $buf = join "", <FILE>; # don't do this at home
  close FILE;

  # parse the data
  my $parsed = GT::HTML->parse( $buf );
  print Dumper( $parsed );


=head2 GT::HTML OO method

  use GT::HTML;
  use GT::URI;
  use GT::Dumper;

  # create the new object
  $parser = new GT::HTML;

  # get some data
  $docs = GT::URI->get( "http://www.gossamer-threads.com/", "http://www.google.com/", "http://www.somethingelse.com" );

  # parse the sucker
  $parser->parse_documents( $docs );
  $parsed = $parser->documents();

  # clear the document list since we're done parsing
  $parser->clear_documents();

  print Dumper($parsed);

=head2 The data structure returned

Just like GT::URI, GT::HTML loves to dabble in complex datastructures. In truth, since GT::HTML was designed to work with GT::URI, 

=head1 METHOD LIST

  sub clear_documents ()  Clears the document cache
  sub de_htmlize ()       decode HTML encoded text
  sub htmlize ()          HTML encode text
  sub parse ()            simple interface to parse out html
  sub parse_documents ()  document parser, caches and returns results

=head2 clear_documents ()

As an object, when an html document is parsed, it is cached within the object for future reference. 

If the object is reused many many times, the cache could become a memory leak!

=head2 de_htmlize ( encoded_text STRING ) : decoded_text STRING

Decodes a HTML encoded string.

=head2 htmlize ( raw_text STRING ) : encoded_text STRING

HTML encodes a raw string.

=head2 GT::HTML->parse ( documents HASHREF/STRING, [ options HASHREF ] ) : parsed_documents HASHREF

Simplest way to parse out an HTML document.

The documents can be simply a buffer with the html document or a hashref in the format of GT::URI's resource data structure.

=head2 parse_documents ( documents HASH/HASHREF/STRING ) : parsed_documents HASHREF

This function was designed to handle the output from GT::URI. In a pinch, simply dump in the document and the parser will assume that the buffer is a text/html document able to be parsed.

=head1 COPYRIGHT

Copyright (c) 2000 Gossamer Threads Inc.  All Rights Reserved.
http://www.gossamer-threads.com/

=head1 VERSION

Revision: $Id: HTML.pm,v 1.13 2002/04/19 17:29:38 aki Exp $

=cut

