# ====================================================================
# Gossamer Threads Module Library - http://gossamer-threads.com/
#
#   GT::HTML::LinkExtractor
#   Author: Scott Beck
#   $Id: LinkExtractor.pm,v 1.6 2002/04/12 21:43:23 sbeck Exp $
#
# Copyright (c) 1999,2000 Gossamer Threads Inc.  All Rights Reserved.
# ====================================================================
#
# Description:
#   A module to extract links from html
#

package GT::HTML::LinkExtractor;

use vars qw/@URL_ALL $ATTRIBS/;
use strict;
use bases 'GT::Base' => '';

use GT::HTML::Parser;

@URL_ALL = qw(
    ftp://
    http://
    https://
    telnet://
    gopher://
    prospero://
    afs://
    file://
    wais:
    mailto:
    nntp:
    news:
);

$ATTRIBS = {
    wanted_urls   => [@URL_ALL],
    wanted_images => ['http://', 'https://'],
    get_images    => 1,
    get_text      => 1,
    follow_base   => 1,
    base          => '',
    links         => []
};

sub extract_from_uri {
# ------------------------------------------------------------------
    my ( $self, @uri ) = @_;
    require GT::URI;
    my $docs = GT::URI->get( @uri );
    my %links;
    for ( keys %$docs ) {
        $self->{base} = $_;
        $links{$_} = $self->extract( $docs->{$_}{buffer} );
    }
    return wantarray ? %links : \%links;
}

sub extract_from_filehandle {
# ------------------------------------------------------------------
    my ( $self, $fh ) = @_;
    read $fh, my $html, -s $fh;
    seek $fh, 0, 0;
    return $self->extract( $html );
}

sub extract_from_file {
# ------------------------------------------------------------------
    my ( $self, $file ) = @_;
    open FH, $file or $self->fatal( 'READOPEN', $file, $! );
    if ( wantarray ) {
        my @ret = $self->extract_from_filehandle( \*FH );
        close FH;
        return @ret;
    }
    else {
        my $ret = $self->extract_from_filehandle( \*FH );
        close FH;
        return $ret;
    }
}

sub extract {
# ------------------------------------------------------------------
    my ( $self, $string ) = @_;
    $self->{base} =~ s,/$,,;
    $self->set_events unless $self->{events_set};
    my %events = ( start_a => $self->{events}{start_a}, end_a => $self->{events}{end_a} );
    if ( $self->{events}{start_base} ) {
        $events{start_base} = $self->{events}{start_base};
    }
    my $parser = new GT::HTML::Parser( events => \%events );
    $parser->parse( $string );
    return wantarray ? @{$self->{links}} : $self->{links};
}

sub links {
# ------------------------------------------------------------------
    my ( $self ) = @_;
    return wantarray ? @{$self->{links}} : $self->{links};
}

sub in_wanted {
# ------------------------------------------------------------------
    my ( $self ) = @_;
    return $self->{in_wanted};
}

sub set_events {
# ------------------------------------------------------------------
    my ( $self ) = @_;
    my $events = {
        start_a => sub {
            my $parser = shift;
            my $tags = pop()->{tags};
            return unless exists $tags->{href};
            if ( defined $tags->{href} and defined $tags->{href}[0] ) {
                $tags->{href}[0] = $self->_add_base( $tags->{href}[0], $self->{wanted_urls} );
            }
            my $match;
            for ( @{$self->{wanted_urls}} ) {
                if ( index( $tags->{href}[0], $_ ) == 0 ) {
                    $match = 1;
                    last;
                }
            }
            return if !$match;
            $parser->handler( start_img => $self->{events}{start_img} ) if $self->{get_images};
            $parser->handler( text => $self->{events}{text} )           if $self->{get_text};
            $self->{in_wanted}++;
            delete $tags->{content};
            push @{$self->{links}}, { map { $_ => $tags->{$_}[0] } keys %$tags };
        },
        end_a => sub {
            my $parser = shift;
            $parser->delete_handler( 'start_img' ) if $self->{get_images};
            $parser->delete_handler( 'text' )      if $self->{get_text};
            $self->{in_wanted}-- unless $self->{in_wanted} == 0;
        }
    };
    if ( $self->{get_images} ) {
        $events->{start_img} = sub {
            my $tags = pop()->{tags};
            return unless exists $tags->{src}[0];
            $tags->{src}[0] = $self->_add_base( $tags->{src}[0], $self->{wanted_images} );
            my $match;
            for ( @{$self->{wanted_images}} ) {
                if ( index( $tags->{src}[0], $_ ) == 0 ) {
                    $match = 1;
                    last;
                }
            }
            return if !$match;
            if ( exists $self->{links}[$#{$self->{links}}]{content} ) {
                if ( ref $self->{links}[$#{$self->{links}}]{content} ) {
                    push @{$self->{links}[$#{$self->{links}}]{content}},  { map { $_ => $tags->{$_}[0] } keys %$tags };
                }
                else {
                    my $data = $self->{links}[$#{$self->{links}}]{content};
                    $self->{links}[$#{$self->{links}}]{content} = [];
                    push @{$self->{links}[$#{$self->{links}}]{content}}, $data,  { map { $_ => $tags->{$_}[0] } keys %$tags };
                }
            }
            else {
                $self->{links}[$#{$self->{links}}]{content} = [  { map { $_ => $tags->{$_}[0] } keys %$tags } ];
            }
        };
    }
    if ( $self->{get_text} ) {
        $events->{text} = sub {
            if ( exists $self->{links}[$#{$self->{links}}]{content} ) {
                if ( ref $self->{links}[$#{$self->{links}}]{content} ) {
                    push @{$self->{links}[$#{$self->{links}}]{content}}, pop()->{data};
                }
                else {
                    my $data = $self->{links}[$#{$self->{links}}]{content};
                    $self->{links}[$#{$self->{links}}]{content} = [];
                    push @{$self->{links}[$#{$self->{links}}]{content}}, $data, pop()->{data};
                }
            }
            else {
                $self->{links}[$#{$self->{links}}]{content} = pop()->{data};
            }
        };
    }
    if ( $self->{follow_base} ) {
        $events->{start_base} = sub {
            my $href = pop()->{tags}{href}[0] || return;
            $self->{base} = $href;
        };
    }
    $self->{events} = $events;
    $self->{events_set} = 1;
    return $self->{events};
}

sub events {
# ------------------------------------------------------------------
    return $_[0]->{events};
}

sub _add_base {
# ------------------------------------------------------------------
    my ( $self, $url, $wanted ) = @_;
    return unless $url;
    return  _dereference_urls( $self->{base}, $url );
}

sub _dereference_urls {
#-------------------------------------------------------------------------------
# takes a url and makes it complete
# assumes HTTP/1.0
    my $base_uri    = shift;
    my $uri         = shift;
    my @deparsed    = ();

# deparse the base host
    $base_uri   =~ m,\w+://(.*)\??,i;
    my $dirs    = _fragment_path( $1 );
    my $host    = shift @{$dirs};
    my $page    = pop @{$dirs};

# now start parsing each of the urls

    if ( $uri =~ m,^\s*\w+:,  ) {
        return $uri;
    }

    else {
        my $rec_path = _fragment_path( $uri );
        my $new_path = $rec_path->[0] ? _merge_paths( $dirs, $rec_path ) : $rec_path;
        return "http://$host/" . join('/', @{_clean_path($new_path) || [] } ) || '';
    }
}

sub _fragment_path {
#-------------------------------------------------------------------------------
# fragments into '/' spaced stuff
    my $path    = shift;

    my @dirs    = ();
    my $i       = 0;
    my $j       = 0;

    while ( ( $i = index( $path, '/', $i ) ) >= 0 ) {
        my $tmp = substr( $path, $j, ($i-$j) );
        push @dirs, $tmp;
        $i++;
        $j = $i;
    }

    push @dirs, substr( $path, $j );

    return \@dirs;
}

sub _clean_path {
#-------------------------------------------------------------------------------
# makes sure that we don't get crap like ///// going on
    my $path    = shift or return undef;

    my @dirs    = ();
    my $page    = pop @{$path};
    @dirs       = grep $_, @{$path};
    push @dirs, $page;
    return \@dirs;
}

sub _merge_paths {
#-------------------------------------------------------------------------------
# merges two paths together
    my $base    = shift;
    my $new     = shift;

# now for each of the new url bases, simulate the file system, back up one, don't do anything, etc
    my @dirs    = @{$base};
    foreach my $dir ( @{$new} ) {

        if ( $dir eq '..' ) {
            pop @dirs;
        }

        elsif ( $dir eq '.' ) {
        }

        else {
            push @dirs, $dir;
        }
    
    }

    return \@dirs;

}

1;

__END__

=head1 NAME

GT::HTML::LinkExtractor - Simple module to extract links from html

=head1 SYNOPSIS

    use GT::HTML::LinkExtractor;
    my $extractor = new GT::HTML::LinkExtractor(
        wanted_urls   => [@GT::HTML::LinkExtractor::URL_ALL],
        wanted_images => ['http://', 'https://'],
        get_images    => 1,
        get_text      => 1,
        follow_base   => 1,
        base          => 'http://somedomain.com'
    );
    $extractor->extract( $html ); # Extract from a string
    $extractor->extract_from_file( "/path/to/html.htm" ); # extract from file
    $extractor->extract_from_uri( "http://www.yahoo.com", "http://www.gossamer-threads.com" ); # extract from uri

    my $links = $extractor->links;

=head1 DESCRIPTION

GT::HTML::LinkExtractor provides simple methods to get back a data structure
of extracted links. It has support of getting images and or text inbetween
<a> </a> tags. The data structure will look like this:

    <a href="http://link" class="MyClass">text<img src="http://image" alt="My Alt Tag">more text</a>

Would be:

    {
        href  => "http://link",
        class => "MyClass"
        content => [
            "text",
            {
                src   => "http://image",
                alt   => "My Alt Tag,
            },
            "more text"
        ]
    }


A common example is:

    <a href="http://www.yahoo.com">Yahoo</a>

Which would be:

    {
        href    => "http://www.yahoo.com",
        content => "Yahoo"
    }



If there is only text, the special key content will be the test, otherwise
it will be an array reference. Any images in the array reference will be
a hash ref of attributes for the image. Any text will be a scalar.

Each one of these data structures is an element in the array ref extract
returns.

=head2 METHODS

GT::HTML::LinkExtractor only has 4 public method for now.

=head2 extract

This method simply takes a string that should be an html page and return
an array ref of links in scalar context and an array of links in list context
see the DESCRIPTION section to see what the structure of a link is.

=head2 extract_from_filehandle

This method takes a file handle returns as extract()

=head2 extract_from_file

This method takes a filepath and returns as extract(). A fatal error is thrown
if the file can not be opened.

=head2 extract_from_url

This method takes a list of uri's and uses the URI module ( see L<GT::URI> ) 
to get the page. Each link that is retrieved in this way set the base path
for that extraction to the link.

This method returns different from extract(), it return a hash in list context
and a hash ref in scalar context. The hash contains the URI address as the key
and the array ref of links as the value.

=head1 OPTIONS

This is a list of options that may either be passed into new() as keys to
a hash, or may be set as calles to methods.

=head2 wanted_urls

This option is an array ref of url types you will to get for links. The link is
tested for this type like:

    if ( index( $link, $type ) == 0 ) {
        $match = 1;
    }

This allows you to weed out thinks like javascript:. By default all types
are supported except javascript:, the types are:

    ftp://
    http://
    https://
    telnet://
    gopher://
    prospero://
    afs://
    file://
    wais:
    mailto:
    nntp:
    news:

=head2 wanted_images

This is similar to wanted_urls in that the test is done the same way. This is
used for image src attributes, useful for skipping images with mailto: as the
source. The default for this is:

    http://
    https://

=head2 get_images

This is a boolian value as to weather or not we are collecting images between
<a> and </a> tags. Default is true.

=head2 get_text

This is a bollian value as to weather or net we are collecting text between
<a> and </a> tags. Default is true.

=head2 base

This is the base url to append to relative urls. Default is empty string.
If you call extract_from_uri() base will be set to the uri.

=head2 follow_base

This is a boolian value. If true when we come accross a <base> tag we will
change the 'base' value to the 'href' in that tag, if one exists. Default
true.

=head1 COPYRIGHT

Copyright (c) 2001 Gossamer Threads Inc.  All Rights Reserved.
http://www.gossamer-threads.com/

=head1 VERSION

Revision: $Id: LinkExtractor.pm,v 1.6 2002/04/12 21:43:23 sbeck Exp $

=cut

