# ====================================================================
# Gossamer Threads Module Library - http://gossamer-threads.com/
#
#   GT::HTML::Parser
#   Author: Scott Beck
#   $Id: Parser.pm,v 1.24 2004/01/13 01:35:17 jagerman Exp $
#
# Copyright (c) 2004 Gossamer Threads Inc.  All Rights Reserved.
# ====================================================================
#
# Description:
#   A module to parse html into a data structure
#

package GT::HTML::Parser;

use strict;
use vars qw($ERRORS $VERSION);
use bases 'GT::Base' => '';

use GT::Text::Tools;

$VERSION = sprintf "%d.%03d", q$Revision: 1.24 $ =~ /(\d+)\.(\d+)/;

sub new {
# ------------------------------------------------------------------
    my ( $class, %opts ) = @_;
    my $self = bless { events => {} }, ref $class || $class;
    if ( $opts{events} ) {
        for ( keys %{$opts{events}} ) {
            $self->handler( $_ => $opts{events}{$_} );
        }
    }
    $self->{remove_space} = exists( $opts{remove_space} ) ? $opts{remove_space} : 0;
    $self->{send_skip}    = exists( $opts{send_skip} )    ? $opts{send_skip}    : 0;
    return $self;
}

sub parse {
# ------------------------------------------------------------------
#
    my ( $self, @html ) = @_;
    my ( $event, $data );
    for my $html ( @html ) {
        if ( ref( $html ) eq 'HASH' ) {
            for ( keys %{$html} ) {
                if ( $self->{events}{document_start} ) {
                    $event = 'document_end';
                    $data  = {
                        name     => 'document_start',
                        uri      => $_,
                        document => $html
                    };
                }
                elsif ( $self->{events}{default} ) {
                    $event = 'default';
                    $data  = {
                        name     => 'document_start',
                        uri      => $_,
                        document => $html
                    };
                }
                if ( $event ) {
                    my $ret = $self->_call_event( $event, $data ) or return;
                    next if $ret < 1;
                }
                $self->parse_single( $html->{$_}{buffer} );
                if ( $self->{events}{document_end} ) {
                    $event = 'document_end';
                    $data  = {
                        name     => 'document_end',
                        uri      => $_,
                        document => $html
                    };
                }
                elsif ( $self->{events}{default} ) {
                    $event = 'default';
                    $data  = {
                        name     => 'document_end',
                        uri      => $_,
                        document => $html
                    };
                }
                if ( $event ) {
                    my $ret = $self->_call_event( $event, $data ) or return;
                    next if $ret < 1;
                }
            }
        }
        else {
            if ( $self->{events}{document_start} ) {
                $event = 'document_end';
                $data  = {
                    name     => 'document_start',
                    document => $html
                };
            }
            elsif ( $self->{events}{default} ) {
                $event = 'default';
                $data  = {
                    name     => 'document_start',
                    document => $html
                };
            }
            if ( $event ) {
                my $ret = $self->_call_event( $event, $data ) or return;
                next if $ret < 1;
            }
            $self->parse_single( $html );
            undef $event;
            if ( $self->{events}{document_end} ) {
                $event = 'document_end';
                $data  = {
                    name     => 'document_end',
                    document => $html
                };
            }
            elsif ( $self->{events}{default} ) {
                $event = 'default';
                $data  = {
                    name     => 'document_end',
                    document => $html
                };
            }
            if ( $event ) {
                my $ret = $self->_call_event( $event, $data ) or return;
                next if $ret < 1;
            }
        }
    }
    return 1;
}

sub _call_event {
    my ( $self, $event, $data ) = @_;
    return -1 if $self->{events}{$event} eq 'SKIP';
    return if $self->{events}{$event} eq 'STOP';
    my @args = ( @{$self->{events}{$event}} > 1 ? @{$self->{events}{$event}}[ 1 .. $#{$self->{events}{$event}} ] : () );
    push @args, $self, $data;
    $self->{events}{$event}[0]->( @args );
    return 1;
}

sub handler {
# ------------------------------------------------------------------
    my ( $self, $event, $handler ) = @_;
    if ( ref( $handler ) eq 'CODE' ) {
        $handler = [ $handler ];
    }
    $self->{events}{$event} = $handler;
}

sub delete_handler {
# ------------------------------------------------------------------
    my ( $self, $event ) = @_;
    delete $self->{events}{$event};
}

sub eof {
# ------------------------------------------------------------------
    $_[0]->{eof} = 1;
}

sub parse_single {
# ------------------------------------------------------------------
    my ( $self, $html ) = @_;

    my $ret = '';
    my $last_pos = 0;
    my $safety = 0;
    my @nodes  = ();
    my ( $last, $pre, $textarea, $skip );
    while ( 
        !$self->{eof} and
        $html =~ m{
            ( # $1 the whole tag
                <
                (
                    ?:
                    !--
                    ( # $2 the attributes are all the data between
                        .*?
                    )
                    --
                    | # or
                    (
                        ?:
                        ( # $3 the name of the tag
                            /?\S+?\b
                        )
                        ( # $4 the attributes
                            [^'">]*
                            (
                                ?:
                                ( # $5 just to match quotes
                                    ['"]
                                )
                                .*?\5
                                [^'">]*
                            )*
                        )
                    )
                )
                >
            )
        }gsx
    )
    {
        my ( $full_tag, $tag, $tag_len, $attribs, @events, @defaults );
        if ( $2 ) {
            ( $full_tag, $tag, $tag_len, $attribs ) = ( $1, '!--', length( $1 ), $2 );
        }
        else {
            ( $full_tag, $tag, $tag_len, $attribs ) = ( $1, lc( $3 ), length( $1 ), $4 );
        }
        my $text_start = $last_pos;
        $last_pos      = pos( $html );

# The text from the last tag to this one
        my $text = substr( $html, $text_start, $last_pos - $tag_len - $text_start );

# Add the text node
        if ( $tag eq '/script' ) {
            if ( $self->{events}{script_guts} ) {
                push @events, [
                    'script_guts', {
                        name    => 'script_guts',
                        data    => $text,
                    }
                ];
            }
            elsif ( $self->{events}{default} ) {
                push @defaults, [
                    'script_guts', {
                        name    => 'script_guts',
                        data    => $text,
                    }
                ]; 
            }
        }
        elsif ( $tag eq '/style' ) {
            if ( $self->{events}{style_guts} ) {
                push @events, [
                    'style_guts', {
                        name    => 'style_guts',
                        data    => $text,
                    }
                ];
            }
            elsif ( $self->{events}{default} ) {
                push @defaults, [
                    'style_guts', {
                        name    => 'style_guts',
                        data    => $text,
                    }
                ];
            }
        }
        elsif ( length $text ) { 
            if ( !$pre and !$textarea and $self->{remove_space} ) {
                $text =~ s/\r?\n/ /g;
                $text =~ s/\s*\Z//;
                $text =~ s/\A\s*//;
                $text =~ s/([\t ])+/$1/g;
            }

            if ( length( $text ) and $text =~ /\S/ ) {
                if ( $skip and $skip eq 'text' ) {
                    $skip = undef;
                    if ( $self->{send_skip} ) {
                        push @events, [
                            'text', {
                                name    => 'text',
                                data    => $text,
                            }
                        ];
                    }
                }
                elsif ( !$skip ) {
                    if ( $self->{events}{text} ) {
                        push @events, [
                            'text', {
                                name    => 'text',
                                data    => $text,
                            }
                        ];
                    }
                    elsif ( $self->{events}{default} ) {
                        push @defaults, [
                            'text', {
                                name    => 'text',
                                data    => $text,
                            }
                        ];
                    }
                }
            }
            elsif ( length( $text ) ) {
                if ( $skip and $skip eq 'space' ) {
                    $skip = undef;
                    if ( $self->{send_skip} ) {
                        push @events, [
                            'space', {
                                name    => 'space',
                                data    => $text,
                            }
                        ];
                    }
                }
                elsif ( !$skip ) {
                    if ( $self->{events}{space} ) {
                        push @events, [
                            'space', {
                                name    => 'space',
                                data    => $text,
                            }
                        ];
                    }
                    elsif ( $self->{events}{default} ) {
                        push @defaults, [
                            'space', {
                                name    => 'space',
                                data    => $text,
                            }
                        ];
                    }
                }
            }
        }

# Don't want to reformat text if we are in a pre tag
        if ( $tag eq 'pre' ) {
            $pre = 1;
        }
        elsif ( $tag eq '/pre' ) {
            $pre = 0;
        }
        elsif ( $tag eq 'textarea' ) {
            $textarea = 1;
        }
        elsif ( $tag eq '/textarea' ) {
            $textarea = 0;
        }

# Start of script or style tag, jump to the end.
        if ( $tag eq 'script' ) {
            pos( $html ) = index( lc( $html ), '</script', $last_pos );
        }
        elsif ( $tag eq 'style' ) {
            pos( $html ) = index( lc( $html ), '</style', $last_pos );
        }

# A comment is special, it can have embedded < and >
        if ( $tag eq '!--' ) {
            if ( $skip and $skip =~ /^(?:skip_)?comment$/ ) {
                if ( $self->{send_skip} ) {
                    push @events, [
                        $skip, {
                            name    => $skip,
                            data    => $attribs,
                        }
                    ];
                }
                $skip = undef;
            }
            elsif ( !$skip ) {
                if ( $self->{events}{comment} ) {
                    push @events, [
                        'comment', {
                            name    => 'comment',
                            data    => $attribs,
                        }
                    ];
                }
                elsif ( $self->{events}{skip_comment} ) {
                    push @events, [
                        'skip_comment', {
                            name       => 'comment',
                            data       => $attribs,
                            event_type => 'skip_comment',
                        }
                    ];
                }
                elsif ( $self->{events}{default} ) {
                    push @defaults, [
                        'comment', {
                            name    => 'comment',
                            data    => $attribs,
                        }
                    ];
                }
            }
        }

# Else we have a non-special tag
        else {
            my %attribs;
            if (
                (
                    index( $tag, '/' ) == -1 and
                    $self->{events}{'start_'.$tag} 
                ) or (
                    index( $tag, '/' ) == 0 and
                    $self->{events}{'end_'.$tag}
                ) or
                $self->{events}{default}
            ) {
                my $pos = 0;
                for ( GT::Text::Tools->linesplit( '\s+' => $attribs ) ) {
                    next unless defined and length;
                    my ( $key, $value ) = GT::Text::Tools->linesplit( '\s*=\s*', $_ );
                    if ( defined $value ) {
                        $key = lc $key;
                        $value =~ s/\A(["'])// and $value =~ s/$1\Z//;
                    }
                    $attribs{$key} = [ $value, $pos ];
                    $pos++;
                }
            }
          
            if ( index( $tag, '/' ) == 0 ) {
                substr( $tag, 0,  1 ) = '';
                if ( $skip and $skip eq "end_$tag" ) {
                    $skip = undef;
                    if ( $self->{send_skip} ) {
                        push @events, [
                            'end_'.$tag, {
                                name       => $tag,
                                tags       => \%attribs,
                                full_tag   => $full_tag,
                                event_type => 'end'
                            }
                        ];
                    }
                }
                elsif ( !$skip ) {
                    if ( $self->{events}{'end_'.$tag} ) {
                        push @events, [
                            'end_'.$tag, {
                                name       => $tag,
                                tags       => \%attribs,
                                full_tag   => $full_tag,
                                event_type => 'end'
                            }
                        ];
                    }
                    elsif ( $self->{events}{'skip_end_'.$tag} ) {
                        push @events, [
                            'skip_end_'.$tag, {
                                name       => $tag,
                                tags       => \%attribs,
                                full_tag   => $full_tag,
                                event_type => 'skip_end'
                            }
                        ];
                    }
                    elsif ( $self->{events}{default} ) {
                        push @defaults, [
                            'end_'.$tag, {
                                name       => $tag,
                                tags       => \%attribs,
                                full_tag   => $full_tag,
                                event_type => 'end'
                            }
                        ];
                    }
                }
            }
            else {
                if ( $skip and $skip eq 'start_'.$tag ) {
                    $skip = undef;
                    if ( $self->{send_skip} ) {
                        push @events, [
                            'start_'.$tag, {
                                name       => $tag,
                                tags       => \%attribs,
                                full_tag   => $full_tag,
                                event_type => 'start'
                            }
                        ];
                    }
                }
                elsif ( !$skip ) {
                    if ( $self->{events}{'start_'.$tag} ) {
                        push @events, [
                            'start_'.$tag, {
                                name       => $tag,
                                tags       => \%attribs,
                                full_tag   => $full_tag,
                                event_type => 'start'
                            }
                        ];
                    }
                    elsif ( $self->{events}{'skip_start_'.$tag} ) {
                        push @events, [
                            'skip_start_'.$tag, {
                                name       => $tag,
                                tags       => \%attribs,
                                full_tag   => $full_tag,
                                event_type => 'skip_start'
                            }
                        ];
                    }
                    elsif ( $self->{events}{default} ) {
                        push @defaults, [
                            'start_'.$tag, {
                                name       => $tag,
                                tags       => \%attribs,
                                full_tag   => $full_tag,
                                event_type => 'start'
                            }
                        ];
                    }
                }
            }
        }
        for ( @events ) {
            my ( $event, $data ) = @$_;
            next if $self->{events}{$event} eq 'SKIP';
            return if $self->{events}{$event} eq 'STOP';
            my @args = ( @{$self->{events}{$event}} > 1 ? @{$self->{events}{$event}}[ 1 .. $#{$self->{events}{$event}} ] : () );
            push @args, $self, $data;
            if ( $data->{event_type} and substr($data->{event_type}, 0, 5) eq 'skip_' ) {
                $skip = $self->{events}{$event}[0]->( @args );
                last if $skip;
                substr($event, 0, 5) = '';
            }
            $self->{events}{$event}[0]->( @args ) if exists $self->{events}{$event};
        }
        for ( @defaults ) {
            my ( $event, $data ) = @$_;
            next if $self->{events}{default} eq 'SKIP';
            return if $self->{events}{default} eq 'STOP';
            my @args = ( @{$self->{events}{default}} > 1 ? @{$self->{events}{default}}[ 1 .. $#{$self->{events}{default}} ] : () );
            push @args, $self, $data;
            $self->{events}{default}[0]->( @args );
        }
    }
    return if $self->{eof};

# It really isn't legal to have text at the end of html...
    my $text = substr( $html, $last_pos );
    if ( defined $text and length $text ) {
        $text =~ s/\A\n+//;
        $text =~ s/\n+\Z//;
        $text =~ s/\n/ /g;
        $text =~ s/([\t ])+/$1/g;
    }

    if ( length $text ) { 
        if ( exists $self->{events}{text} ) {
            unless ( $self->{events}{text} eq 'SKIP' or $self->{events}{text} eq 'STOP' ) {
                $self->{events}{text}[0]->(
                    ( @{$self->{events}{text}} > 1 ? @{$self->{events}{text}}[ 1 .. $#{$self->{events}{text}} ] : () ),
                    $self,
                    {
                        name    => 'text',
                        data    => $text,
                    }
                );
            }
        }
        elsif ( exists $self->{events}{default} ) {
            unless ( $self->{events}{default} eq 'SKIP' or $self->{events}{default} eq 'STOP' ) {
                $self->{events}{default}[0]->(
                    ( @{$self->{events}{default}} > 1 ? @{$self->{events}{default}}[ 1 .. $#{$self->{events}{default}} ] : () ),
                    $self,
                    {
                        name    => 'text',
                        data    => $text,
                    }
                );
            }
        }
    }
}

sub tag_to_string {
# ------------------------------------------------------------------
    my ( $self, $tag ) = @_;
    my $end;
    my $ret;
    if ( $tag->{event_type} eq 'end' ) {
        return $ret .'</'.$tag->{name}.'>';
    }
    $ret = '<'.$tag->{name}.' ';
    $ret .= '/' if $tag->{event_type} eq 'end';
    $end = '/' if delete $tag->{tags}{'/'};
    for ( sort { $tag->{tags}{$a}[1] <=> $tag->{tags}{$b}[1] } keys %{$tag->{tags}} ) {
        if ( !defined( $tag->{tags}{$_}[0] ) ) {
            $ret .= $_.' ';
        }
        else {
            $ret .= $_.'='.'"'.$tag->{tags}{$_}[0].'" ';
        }
    }
    chop $ret;
    $ret .= $end if $end;
    $ret .= '>';
    return $ret;
}

1;

__END__

=head1 NAME

GT::HTML::Parser - Event driven html parser

=head1 SYNOPSIS

    my $parser = new GT::HTML::Parser(
        events => {
            document_start => \&handle_start,
            document_end   => \&handle_end,
            comment        => \&handle_comments,
            text           => [ \&handle, $arg1, $arg2 ],
            start_img      => 'SKIP',
            start_a        => \&handle_start_link,
            end_a          => \&handle_end_link
        }
    );
    my $documents = GT::URI->get( "http://www.gossamer-threads.com", "http://www.yahoo.com", "http://www.google.com" );
    $parser->parse( $documents );

=head1 DESCRIPTION

GT::HTML::Parser is an event driven html parser. It is very close to a generic SGML parser class
but understands things like javascript.

=head2 new

This is the constructor method. Takes a hash of options.

The 'events' key should be a hash reference
of events. The keys in the hash reference are one of

    document_start   - Fired at the start of document parsing, good for when you are parsing multiple
                       documents.
    document_end     - Fired when a document is finished being parsed, again only useful for parsing
                       multiple documents.
    comment          - Fired when a comment tag is encountered.
    skip_comment     - Same as skip_start_<tag> (see below).
    text             - Fired when text that is not an html tag is encountered, be careful this includes
                       text inside things like <textarea> and <option> lists.
    space            - Fired when something outside of tags is nothing but space, the text event will
                       not be called, this event will. Space is \s+
    script_guts      - Fired when a scripting languages is encountered, this does not include the start
                       and end tag of the script.
    style_guts       - Like script_guts but for <style> guts.
    start_<tag>      - Where <tag> is any html tag. Fired when that start of that tag is encountered. e.g.
                       no / at the start.
    skip_start_<tag> - Like start_<tag>, except it can return the name of an event (ie. start_a). Until
                       that event (start_a) is encountered, no other events will be fired.  By default,
                       that event will not be fired.  If you want the event to be fired, use the send_skip
                       option.
    end_<tag>        - Where <tag> is any html tag. Fired when the end of that tag is encountered. e.g.
                       a / at the start.
    skip_end_<tag>   - Just like skip_start_<tag>, except it works on closing tags.
    default          - This is fired for any event that does not have an event set, even tag events.
                       (be careful, setting this will really slow down parsing)

These are all the valid events. The values to these keys may be one of three things. 

An array reference where the first element must be a code ref. The rest of the elements will be passed 
to the code ref as it's first argument.

A code ref, which will be called with normal event arguments (see below).

A processor command, one of:
    SKIP
    STOP

SKIP is the same as setting the callback as an empty code ref, useful if you set a 
default event and want to skip some tags. The is less overhead than an empty code
ref.

STOP stops processesing the html and returns.

Use 'remove_space' if you want text normalized before calling your text event. This is any excess
white space will be removed from the text.

Use 'send_skip' if you want it to fire events on skip_*, then use this, otherwise the skip event will
not be fired.

=head2 handler

This is a method to setup an event handler. The first argument is the event to set the handler for,
the second argument either a code reference and array reference or a scalar, just like calling new().
For example to set up a handler to skip <em> you would do:

    $parser->handler( start_em => 'SKIP' );

=head2 delete_handler

This is a method to remove an event handler. The only argument is the event you wish to remove, this
is usually called from within an event to skip things.

=head2 parse

This is the only other method in this module that is public. This method starts the parsing of the docuemt(s)
specified at the arguments. Any combination of scalar (containing the html) or return from GT::URI->get()
may be passed in. This method parsed the document(s) and calles your events.

Events argument are by default: 
    1. The parser
    2. Hash reference of data pertaining to the event.

You may add arguments in front of these by passing an array reference as the event to new(),
(see new above).

The hash reference will be slightly different per event. Here is a reference of event to structure:

    Event Type        Key        Value
    ------------------------------------------------------------------------------------------------
    document_start - 
                      name     - the name of the event 'docuemnt_start'
                      uri      - the uri of the document. Note: uri only exists if you passed in the 
                                 return of GT::URI->get()
                      document - the entire html document
                     
    document_end   - 
                      name     - the name of the event 'docuemnt_end'
                      uri      - the uri of the document. Note: uri only exists if you passed in the 
                                 return of GT::URI->get()
                      document - the entire html document
                     
    comment        - 
                      name     - the name of the event 'comment',
                      data     - The comments content,
                      
    skip_comment   - 
                      name     - the name of the event 'comment',
                      data     - The comments content,

    text           - 
                      name     - the name of the event 'text',
                      data     - the text
                     
    space          - 
                      name     - the name of the event 'space',
                      data     - the spaces \s+
                      
    script_guts    - 
                      name     - the name of the event 'script_guts',
                      data     - the script
                     
    style_guts     - 
                      name     - the name of the event 'style_guts',
                      data     - the style
                     
    skip_start_<tag>
    start_<tag>    -
                      name     - The name of the tag: for <a href="foo.com"> a would be the name
                      tags     - A hash reference of the attributes e.g. href would be a key
                                 and the url would be the value in a <a> tag
                      tag_type - The type of start or end,  'start' or 'skip_start'.
                      full_tag - The complete tag.
                      
    skip_end_<tag>
    end_<tag>      - 
                      name     - The name of the tag: for </a> a would be the name
                      tags     - A hash reference of the attributes
                      tag_type - The type of start or end,  'end'
                      full_tag - The complete tag.
                    
    default        -  This one is special, it will be called with the arguments of the event is
                      is replacing.

=head2 tag_to_string

This is a simple method that takes the data pertaining to your event and creates the tag from that.

    my $parser = new GT::HTML::Parser(
        events => {
            start_a => sub {
                my ( $parser, $data ) = @_;
                delete $data->{tags}{href};
                print "Removed class from", $parser->tag_to_string( $data );
            }
        }
    );

=head1 EXAMPLES

See L<GT::HTML::LinkExtractor> and See L<GT::HTML::ToText>

=head1 COPYRIGHT

Copyright (c) 2004 Gossamer Threads Inc.  All Rights Reserved.
http://www.gossamer-threads.com/

=head1 VERSION

Revision: $Id: Parser.pm,v 1.24 2004/01/13 01:35:17 jagerman Exp $

=cut

