# ====================================================================
# Gossamer Threads Module Library - http://gossamer-threads.com/
#
#   GT::HTML::ToText
#   Author: Scott Beck
#   $Id: ToText.pm,v 1.11 2004/02/19 18:26:16 sbeck Exp $
#
# Copyright (c) 2004 Gossamer Threads Inc.  All Rights Reserved.
# ====================================================================
#
# Description:
#   A module to convert html to text
#


package GT::HTML::ToText;

use strict;
use vars qw($ERRORS $VERSION);

use bases 'GT::Base' => '';

use GT::HTML::Parser;
use GT::HTML::LinkExtractor;
use GT::Text::Tools;

$ERRORS = {
    HTML_INVALID => 'HTML is not valid enough to finish parsing'
};

$VERSION = sprintf "%d.%03d", q$Revision: 1.11 $ =~ /(\d+)\.(\d+)/;

sub convert {
# ------------------------------------------------------------------
# Internal method to conver html to text.
#
    my ($self, $convert, $opts) = @_;
    $opts ||= {};
    exists( $opts->{linewrap} ) or $opts->{linewrap} = 70;
    $opts->{base} ||= '';
    
    
    my $text = '';
    my ( $start_li, @ol_counter, $last_l, $block_depth, $ul_depth, $ol_depth, $ahref, $ex, $events, $links );
    $events = {};
    unless ( $opts->{no_links} ) {
        $ex = new GT::HTML::LinkExtractor(
            get_images => 0,
            get_text   => 0,
            base       => $opts->{base}
        );
        $events = $ex->set_events;
        $links  = $ex->links;
    }
    my $parser = new GT::HTML::Parser(
        remove_space => 1,
        events => {
            %$events,
            start_hr => sub {
                $text .= "\n";
                my $len = $opts->{linewrap} || 70;
                $text .= '-' x $len;
                $text .= "\n";
            },
            start_li => sub {
                $text  .= "\n";
                if ( $last_l eq 'ol' ) {
                    my $count = $ol_counter[-1];
                    $text .= "$count. ";
                    $ol_counter[-1]++;
                }
                else {
                    $text .= "* ";
                }
                $start_li = 1;
            },
            start_p => sub {
                $text .= "\n\n";
            },
            start_div => sub {
                $text .= "\n";
            },
            end_div => sub {
                $text .= "\n";
            },
            start_blockquote => sub {
                $text .= "\0blockquote".$block_depth++.";\n";
            },
            end_blockquote => sub {
                $text .= "\n\0blockquote".--$block_depth.";";
            },
            start_ol => sub {
                $text  .= "\0ol".$ol_depth++.";";
                $last_l = 'ol';
                push @ol_counter, 1;
            },
            end_ol => sub {
                $text .= "\0ol".--$ol_depth.";";
                pop @ol_counter;
            },
            start_ul => sub {
                $text .= "\0ul".$ul_depth++.";";
                $last_l = 'ul';
            },
            end_ul => sub {
                $text .= "\0ul".--$ul_depth.";";
            },
            start_br => sub {
                $text .= "\n";
            },
            text => sub {
                my $ntext = pop()->{data};
                unless ( $opts->{no_links} ) {
                    if ( $ex->in_wanted ) {
                        $ntext .= " *".@{$links};
                    }
                }
                if ( $text =~ /\s\Z/ ) {
                    $text .= $ntext;
                }
                elsif ( !length $text ) {
                    $text = $ntext;
                }
                else {
                    $text .= ' ' . $ntext;
                }
            }
        }
    );
    $parser->parse( $convert );
    $text =~ s/\A\s*//;
    $text =~ s/\s*\Z//;
    $text =~ s/^[\t ]+//mg;
    $text =~ s/[\t ]+$//mg;
    $text = GT::Text::Tools->linewrap( $text, $opts->{linewrap}, { nowrap => [ 'http://\S+' ] } ) if $opts->{linewrap};
    $text = _rec_string( $text );
    if ($text =~ /[^\x0d\x0a\x20-\x7e]/) {
        $text =~ s/[^\x0d\x0a\x20-\x7e]//g;
        $self->warn( 'HTML_INVALID' );
    }
    $text =~ s/&amp;/&/g;
    $text =~ s/&gt;/>/g;
    $text =~ s/&lt;/</g;
    $text =~ s/&quot;/"/g;
    $text =~ s/&nbsp;/ /g;
    $text =~ s/&#?\w+;//g;
    $text =~ s/^/$opts->{quote}/gm if $opts->{quote};
    if ( !$opts->{no_links} and @$links ) {
        $text .= "\n".join( "\n", map { $_.'. '.$links->[$_ - 1]->{href} } 1 .. @$links )."\n";
    }
    return $text;
}

sub strip {
# ------------------------------------------------------------------
    my ($self, $convert, $opts) = @_;
    $opts ||= {};
    exists( $opts->{linewrap} ) or $opts->{linewrap} = 70;
    $opts->{base} ||= '';
    
    
    my $text;
    my $parser = new GT::HTML::Parser(
        remove_space => 1,
        events => {
            text => sub {
                my $ntext = pop()->{data};
                if ( $text =~ /\s\Z/ ) {
                    $text .= $ntext;
                }
                elsif ( !length $text ) {
                    $text = $ntext;
                }
                else {
                    $text .= ' ' . $ntext;
                }
            }
        }
    );
    $parser->parse( $convert );
    $text =~ s/\A\s*//;
    $text =~ s/\s*\Z//;
    $text =~ s/^[\t ]+//mg;
    $text =~ s/[\t ]+$//mg;
    $text = GT::Text::Tools->linewrap( $text, $opts->{linewrap}, { nowrap => [ 'http://\S+' ] } ) if $opts->{linewrap};
    $text =~ s/&amp;/&/g;
    $text =~ s/&gt;/>/g;
    $text =~ s/&lt;/</g;
    $text =~ s/&quot;/"/g;
    $text =~ s/&nbsp;/ /g;
    $text =~ s/&#?\w+;//g;
    $text =~ s/^/$opts->{quote}/gm if $opts->{quote};
    return $text;
}

sub _rec_string {
# ------------------------------------------------------------------
    my ( $ret ) = @_;
    my $rret = '';
    my $last_pos = 0;
    while ( $ret =~ /(\x00((?:blockquote|ol|ul)\d+);(.*?)\x00\2;)/sg ) {
        my $t = $3;
        my $tag_len      = length($1);
        my $print_start  = $last_pos;
        $last_pos        = pos( $ret );

        # The text from the last tag to this one
        my $text = substr( $ret, $print_start, $last_pos - $tag_len - $print_start );
        if ( index( $t, "\0" ) != -1 ) {
            $t = _rec_string( $t );
        }
        $t =~ s/^/    /mg;
        $rret .= $text."\n".$t."\n";
    }
    $rret .= substr( $ret, $last_pos );
    return $rret;
}

1;

__END__

=head1 NAME

GT::HTML::ToText - Simple module to convert html to text

=head1 SYNOPSIS

    use GT::HTML::ToText;
    my $text = GT::HTML::ToText->convert( $html, { linewrap => 70, quote => '> '  } );

=head1 DESCRIPTION

GT::HTML::ToText provides a single method to covert a string of html to text. It
is mostly useful for email html when you need to reproduce the same html content
as text. This module understands simple blockquotes, ordered list, and unorderedlist
and converts them to the text equiv. No all spacing will be perfect but I tried
to get it as close as I could.

=head2 convert

This is the only public method, there is no new method.
    
    my $text = GT::HTML::ToText->convert( $html, $opts );

The first argument is the html text you want to convert, the second is a hash
reference of options. Options are quote, linewrap, no_links and base.

base is the base url to append to links that are relative.

Set no_links to stop ToText from getting the list of links in the document and
putting them at the end of the text file.

quote is used mainly when replying to html email and you wish it quoted, usually
something like '> '. This will be prepended to the beginning of each line after
indention and line wrap is applied.

linewrap is the width you wish text to be wrapped on, baring indentation. The
default for this is 70, set to a non-true value to disable linewrap.

=head1 COPYRIGHT

Copyright (c) 2004 Gossamer Threads Inc.  All Rights Reserved.
http://www.gossamer-threads.com/

=head1 VERSION

Revision: $Id: ToText.pm,v 1.11 2004/02/19 18:26:16 sbeck Exp $

=cut


