# ==================================================================
# Gossamer Threads Module Library - http://gossamer-threads.com/
#
#   GT::Plugins
#   Author  : Alex Krohn
#   CVS Info :                     
#   $Id: Plugins.pm,v 1.50 2004/08/23 19:54:27 jagerman Exp $
#
# Copyright (c) 2004 Gossamer Threads Inc.  All Rights Reserved.
# ==================================================================
#
# Description: A plugin system for CGI scripts.
#

package GT::Plugins;
# ==================================================================
use strict;
use vars qw/$VERSION $DEBUG $ERRORS %PLUGIN_CFG $ATTRIBS $ACTION $error @ISA $PROFILE $AUTOLOAD/;
use GT::Base;
use GT::Config;
use GT::AutoLoader;

@ISA    = qw/GT::Base/;
$ERRORS = {
    BADARGS      => "Invalid arguments. Usage: %s",
    CANTLOAD     => "Unable to load plugin %s Reason: %s",
    CANTOPEN     => "Unable to open: %s (%s)",
    CANTDELETE   => "Unable to remove plugin file: %s (%s)",
    CANTMOVE     => "Unable to move plugin %s from '%s' to '%s' (%s)",
    CANTREMOVE   => "Unable to remove plugin file %s (%s)",
    PLUGEXISTS   => "The plugin %s already exists, unable to overwrite without confirmation",
    NOINSTALL    => "Unable to load install code in plugin %s. Missing Install.pm file.",
    NOCODE       => "Unable to load main code for plugin %s from tar file. Missing %s.pm file.",
    NOPLUGINNAME => "Please name your plugin before calling save()",
    NOPLUGIN     => "There is no plugin named: %s in the config file.",
    CORRUPTCFG   => "Syntax error in config file: %s",
    PLUGINERR    => "Error running plugin %s hook: %s. Reason: %s"
};
$ATTRIBS = { directory => undef, prefix => '' };
$DEBUG   = 0;
$VERSION = sprintf "%d.%03d", q$Revision: 1.50 $ =~ /(\d+)\.(\d+)/;

# Profile dispatches?
$PROFILE = 0;

# Actions that plugins can handle.
use constants(
    STOP     => 1,
    CONTINUE => 2,

    NAME    => 0,
    TYPE    => 1,
    HOOK    => 2,
    ENABLED => 3
);

sub import {
# --------------------------------------------------------------------
# Exports the STOP and CONTINUE subs.
#
    my $pkg     = shift;
    my $callpkg = caller;

    no strict 'refs';
    *{$callpkg . '::STOP'}      = \&STOP;
    *{$callpkg . '::CONTINUE'}  = \&CONTINUE;
}

sub init {
# -----------------------------------------------------------------
# Set our debug level and any extra options.
#
    my $self = shift;
    $self->set(@_);
    if ($self->{debug}) {
        $self->{_debug} = delete $self->{debug};
    }

    return $self;
}

sub active_plugins {
# -----------------------------------------------------------------------------
# Class/object method that returns a boolean value indicating whether or not
# the given argument (a plugin hook name) has any registered plugin hooks.
# Primarily designed for optimizations where a section of code isn't needed
# except for plugins.
#
    my $self = shift;
    my $directory = ref $self && $self->{directory} || shift;
    my $hook_name = lc shift;
    defined $PLUGIN_CFG{$directory} or $self->load_cfg($directory);

    return (
        (
            exists $PLUGIN_CFG{$directory}->{_pre_hooks}->{$hook_name}
            and @{$PLUGIN_CFG{$directory}->{_pre_hooks}->{$hook_name}}
        )
            or
        (
            exists $PLUGIN_CFG{$directory}->{_post_hooks}->{$hook_name}
            and @{$PLUGIN_CFG{$directory}->{_post_hooks}->{$hook_name}}
        )
            or
        $PROFILE
    ) ? 1 : undef;
}

sub dispatch {
# -----------------------------------------------------------------
# Class Method to Run plugins.
#
    my $self = shift;
    my $directory = ref $self && $self->{directory} ? $self->{directory} : shift;
    my ($hook_name, $code, @args) = @_;
    $hook_name = lc $hook_name;
    defined $PLUGIN_CFG{$directory} or $self->load_cfg($directory);


# Run any pre hooks.
    my @results;
    if (exists $PLUGIN_CFG{$directory}->{_pre_hooks}->{$hook_name}) {
        local $^W; no strict 'refs';
# Save our action in case plugins is called twice.
        my $orig_action = $ACTION;
        foreach my $hook (@{$PLUGIN_CFG{$directory}->{_pre_hooks}->{$hook_name}}) {
            $self->debug("Plugin: pre $hook_name running => $hook") if ref $self ? $self->{_debug} : $DEBUG;
            defined &{$hook} or $self->_load_hook($hook, 'PRE') or next;
            $ACTION = CONTINUE;
            @results = $hook->(@args);
            if ($ACTION == STOP) {
                $self->debug("Plugin pre $hook_name stopped further plugins.") if ref $self ? $self->{_debug} : $DEBUG;
                last;
            }
        }
        unless ($ACTION == STOP) {
            @results = $code->(@args);
        }
        $ACTION = $orig_action;
    }
    else {
        @results = $code->(@args);
    }

# Run any post hooks.
    if (exists $PLUGIN_CFG{$directory}->{_post_hooks}->{$hook_name}) {
        local ($^W); no strict 'refs';
        my $orig_action = $ACTION;
        foreach my $hook (@{$PLUGIN_CFG{$directory}->{_post_hooks}->{$hook_name}}) {
            $self->debug("Plugin: post $hook_name running => $hook") if ref $self ? $self->{_debug} : $DEBUG;
            defined &{$hook} or $self->_load_hook($hook, 'POST') or next;
            $ACTION = CONTINUE;
            @results = $hook->(@results);
            if ($ACTION == STOP) {
                $self->debug("Plugin post $hook_name stopped further plugins.") if ref $self ? $self->{_debug} : $DEBUG;
                last;
            }
        }
        $ACTION = $orig_action;
    }

# Run profiler
    if ($PROFILE) {
        require GT::Plugins::Profiler;
        local $self->{directory} = $directory;
        GT::Plugins::Profiler->profile($self, $hook_name, \@args, \@results);
    }

# Must return as a list, turn off -w as can be undef.
    local $^W;
    return ((@results)[0 .. $#results]);
}

sub dispatch_method {
# -----------------------------------------------------------------
# Class Method to Run plugins.
#
    my $self = shift;
    my $directory = ref $self && $self->{directory} ? $self->{directory} : shift;
    my ($hook_name, $object, $method, @args) = @_;
    $hook_name = lc $hook_name;
    defined $PLUGIN_CFG{$directory} or $self->load_cfg($directory);

# Run any pre hooks.
    my @results;
    if (exists $PLUGIN_CFG{$directory}->{_pre_hooks}->{$hook_name}) {
        local ($^W); no strict 'refs';
# Save our action in case plugins is called twice.
        my $orig_action = $ACTION;
        foreach my $hook (@{$PLUGIN_CFG{$directory}->{_pre_hooks}->{$hook_name}}) {
            $self->debug("Plugin: pre $hook_name running => $hook") if ref $self ? $self->{_debug} : $DEBUG;
            defined &{$hook} or $self->_load_hook($hook, 'PRE') or next;
            $ACTION = CONTINUE;
            @results = $hook->($object, @args);
            $ACTION == STOP and last;
        }
        unless ($ACTION == STOP) {
            @results = $object->$method(@args);
        }
        $ACTION = $orig_action;
    }
    else {
        @results = $object->$method(@args);
    }

# Run any post hooks.
    if (exists $PLUGIN_CFG{$directory}->{_post_hooks}->{$hook_name}) {
        local ($^W); no strict 'refs';
        my $orig_action = $ACTION;
        foreach my $hook (@{$PLUGIN_CFG{$directory}->{_post_hooks}->{$hook_name}}) {
            $self->debug("Plugin: post $hook_name running => $hook") if ref $self ? $self->{_debug} : $DEBUG;
            defined &{$hook} or $self->_load_hook($hook, 'POST') or next;
            $ACTION = CONTINUE;
            @results = $hook->($object, @results);
            $ACTION == STOP and last;
        }
        $ACTION = $orig_action;
    }

# Run profiler
    if ($PROFILE) {
        require GT::Plugins::Profiler;
        ref $self or $self = $self->new;
        $self->{directory} = $directory;
        GT::Plugins::Profiler->profile($self, $hook_name, \@args, \@results);
    }

# Must return as a list, turn off -w as can be undef.
    local $^W;
    return ((@results)[0 .. $#results]);
}

sub load_cfg {
# -----------------------------------------------------------------
# Load the plugin config file.
#
    my $self      = shift;
    my $directory = shift || '.';

    local ($@, $!);
    my $cfg = GT::Config->load("$directory/plugin.cfg", { local => 0, inheritance => 0, create_ok => 1 });

    if (!$cfg and $@ || $! and ref $self ? $self->{_debug} : $DEBUG) {
        $self->debug("Plugin: invalid plugin config file: $directory/plugin.cfg Reason: $@ $!");
    }
    $PLUGIN_CFG{$directory} = $cfg || {};

# Load the hooks for easy access.
    $PLUGIN_CFG{$directory}->{_pre_hooks}  = {};
    $PLUGIN_CFG{$directory}->{_post_hooks} = {};
    foreach my $plugin (keys %{$PLUGIN_CFG{$directory}}) {
        next if (substr($plugin, 0, 1) eq '_');
        next unless (ref $PLUGIN_CFG{$directory}->{$plugin}->{hooks} eq 'ARRAY');
        foreach my $hook (@{$PLUGIN_CFG{$directory}->{$plugin}->{hooks}}) {
            next unless $hook->[ENABLED]; # Make sure it's enabled.
            if ($hook->[TYPE] eq 'PRE') {
                push @{$PLUGIN_CFG{$directory}->{_pre_hooks}->{lc $hook->[NAME]}}, $hook->[HOOK];
            }
            elsif ($hook->[TYPE] eq 'POST') {
                push @{$PLUGIN_CFG{$directory}->{_post_hooks}->{lc $hook->[NAME]}}, $hook->[HOOK];
            }
        }
    }
    return $PLUGIN_CFG{$directory};
}

$COMPILE{save_cfg} = __LINE__ . <<'END_OF_SUB';
sub save_cfg {
# -----------------------------------------------------------------
# Save the plugin cfg file.
#
    my $self = shift;
    my $directory   = shift;
    my $config      = @_ ? shift : $PLUGIN_CFG{$directory};
    my ($pre, $post) = delete @$config{qw/_pre_hooks _post_hooks/};

    $config->save();

    @$config{qw/_pre_hooks _post_hooks/} = ($pre, $post);
    return 1;
}
END_OF_SUB

sub action {
# -------------------------------------------------------------------
# Sets the action the plugin wants.
#
    $ACTION = $_[1];  # $_[0] = $self or GT::Plugins
}

$COMPILE{_load_hook} = __LINE__ . <<'END_OF_SUB';
sub _load_hook {
# -------------------------------------------------------------------
# Loads a module and checks for the hook.
#
    my ($self, $hook, $stage) = @_;
    my ($pkg) = $hook =~ /^(.*)::[^:]+$/ or return;
    $pkg =~ s,::,/,g;
    {
        local $SIG{__DIE__}; eval { require "$pkg.pm"; };
    }
    if ($@) {
        return $self->error('PLUGINERR', 'FATAL', $stage, $hook, "$@");
    }
    if (! defined &{$hook}) {
        return $self->error('PLUGINERR', 'FATAL', $stage, $hook, "$hook does not exist in $pkg");
    }
    return 1;
}
END_OF_SUB

$COMPILE{reset_env} = __LINE__ . <<'END_OF_SUB';
sub reset_env { }
END_OF_SUB

1;

__END__

=head1 NAME

GT::Plugins - a plugin interface for Gossamer Threads products.

=head1 SYNOPSIS

    use GT::Plugins;
    GT::Plugins->dispatch('/path/to/plugin/dir', 'plugin_name', \&code_ref, @args);
    GT::Plugins->dispatch_method('/path/to/plugin/dir', 'plugin_name', $self, 'method', @args);

=head1 DESCRIPTION

The plugin module provides dispatch() the heart of the entire system.
The dispatch function takes a plugin name and does the following:

1. Loads the plugin config file.

2. Runs any registered PRE hooks found in the config file.

2a. If no hooks are found, or hooks found but the hook does not set
    the action to STOP, the code_ref passed in is run.

2b. If a hook sets the action to STOP, then the code ref is not run.

3. Runs any registered POST hooks found in the config file

4. Returns last result.

Also included as part of the plugin system are some modules for
web based tools to manage plugins:

L<GT::Plugins::Manager> - Add, remove and edit plugin files.

L<GT::Plugins::Wizard> - Create shell plugins.

L<GT::Plugins::Installer> - Used in installing plugins.

Setting the $PROFILE variable in the .pm file will cause the module
method to catch dipatches and write information related to their
calling in a new file called hooks.cfg in the same directory as 
plugins.cfg.

=head1 COPYRIGHT

Copyright (c) 2004 Gossamer Threads Inc.  All Rights Reserved.
http://www.gossamer-threads.com/

=head1 VERSION

Revision: $Id: Plugins.pm,v 1.50 2004/08/23 19:54:27 jagerman Exp $

=cut
